/**
 * Copyright (C) 2003-2004 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <time.h>

#include <rapi.h>

#include "com_funambol_syncclient_sps_pocketpc_RapiDB.h"

#include "rapidb.h"
#include "debug.h"
#include "error.h"
#include "cedb.h"


/*
 * DLL entry point
 */
BOOL APIENTRY DllMain( HINSTANCE instance, 
                       DWORD     reason, 
                       LPVOID    reserved
					 )
{
#ifdef DEBUG
    printf("JRAPI loaded!\n");
#endif

    switch(reason) {
        case DLL_PROCESS_ATTACH:
            initialized = FALSE;
            init();
            break;

        case DLL_PROCESS_DETACH:
            uninit();
            break;
    }

    
    return TRUE;
}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    createDB
 * Signature: (II)Z
 */
JNIEXPORT jboolean JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_createDB(JNIEnv *env, jobject o, jint key, jint status) {
    HANDLE   db       = 0    ;
    CEOID    oid      = 0    ;
    CEGUID   volumeId        ;
    WCHAR*   dbName          ;
    WCHAR*   volume          ;
    jboolean ret      = FALSE;


    jstring jstrVolume = NULL;
    jstring jstrDBName = NULL;    


    if (!sanityCheck(env, o, CHECK_RAPI)) {
        //
        // we got an excption
        //
        return FALSE;
    }

    //
    // First mount the volume, then open the database
    //

    jstrVolume = getVolume(env, o);
    jstrDBName = getDBName(env, o);    

    volume = (WCHAR*)(*env)->GetStringChars(env, jstrVolume, NULL);
    dbName = (WCHAR*)(*env)->GetStringChars(env, jstrDBName, NULL);

#ifdef DEBUG
    wprintf(TEXT("Creating database '%s' in %s with key field %d and status field %d\n"), dbName, volume, key, status);
#endif

    memset(&volumeId, 0, sizeof(CEGUID));
    if (mountDBVolume(volume, &volumeId) == FALSE) {
        goto finally;
    }

   
    oid = createDB( &volumeId, 
                    dbName, 
                    MAKELONG(CEVT_LPWSTR, key), 
                    MAKELONG(CEVT_LPWSTR, status));

    if (oid == 0) {
        error(TEXT("Error creating the database"));
        goto finally;
    } else {
        setDBOid(env, o, oid);
        setVolOid(env, o, &volumeId);
        setKeyField(env, o, key);
        setStatusField(env, o, status);
        ret = TRUE;
    }

finally:
    (*env)->ReleaseStringChars(env, jstrVolume, volume);
    (*env)->ReleaseStringChars(env, jstrDBName, dbName);

    return ret;
}


/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    openDB
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_openDB(JNIEnv* env, 
                                                        jobject o  ,
                                                        jint key   ,
                                                        jint status,
                                                        jint sort  ) {
    HANDLE   db       = 0    ;
    CEOID    oid      = 0    ;
    CEGUID   volumeId        ;
    WCHAR*   dbName          ;
    WCHAR*   volume          ;
    jboolean ret      = FALSE;

    jstring jstrVolume = NULL;
    jstring jstrDBName = NULL;    


    if (!sanityCheck(env, o, CHECK_RAPI)) {
        //
        // we got an excption
        //
        return FALSE;
    }

    //
    // First mount the volume, then open the database
    //

    jstrVolume = getVolume(env, o);
    jstrDBName = getDBName(env, o);    

    volume = (WCHAR*)(*env)->GetStringChars(env, jstrVolume, NULL);
    dbName = (WCHAR*)(*env)->GetStringChars(env, jstrDBName, NULL);

#ifdef DEBUG
    wprintf(TEXT("Opening database '%s' in %s with sort order %d\n"), dbName, volume, sort);
#endif

    oid = getDBOid(env, o);
    memset(&volumeId, 0, sizeof(CEGUID));
    if (mountDBVolume(volume, &volumeId) == FALSE) {
        goto finally;
    }
   
    db = openDB(&volumeId, dbName, MAKELONG(CEVT_LPWSTR, sort), &oid);

    if (db == INVALID_HANDLE_VALUE) {
        goto finally;
    } else {
        setDBHandle(env, o, db);
        setDBOid(env, o, oid);
        setVolOid(env, o, &volumeId);
        setKeyField(env, o, key);
        setStatusField(env, o, status);
        setSortField(env, o, sort);
        ret = TRUE;
    }

finally:
    (*env)->ReleaseStringChars(env, jstrVolume, volume);
    (*env)->ReleaseStringChars(env, jstrDBName, dbName);

    return ret;
}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    closeDB
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_closeDB(JNIEnv* env, 
                                                         jobject o  ) {
    CEGUID volumeId;

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return;
    }

#ifdef DEBUG
    wprintf(TEXT("Closing database %d\n"), getDBHandle(env, o));
#endif

    CeCloseHandle(getDBHandle(env, o));
    getVolOid(env, o, &volumeId);
    CeUnmountDBVol(&volumeId);

}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    deleteDB
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_deleteDB(JNIEnv* env, 
                                                          jobject o  ) {
    HANDLE db  = getDBHandle(env, o);
    CEOID  oid = getDBOid(env, o);

    CEGUID guid;

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return;
    }

#ifdef DEBUG
    wprintf(TEXT("Deleting database %d\n"), db);
#endif

    //
    // First of all close the database, otherwise it will result locked
    //

    CeCloseHandle(db);

    getVolOid(env, o, &guid);
    if (CeDeleteDatabaseEx(&guid, oid) == FALSE) {
        WCHAR msg[256];
        
        wsprintf(msg, TEXT("Error deleting the database %d"), db);

        error(msg);
    }
}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    readRecord
 * Signature: (Ljava/lang/String;)[Ljava/lang/Object;
 */
JNIEXPORT jobjectArray JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_readRecord(JNIEnv* env, 
                                                            jobject o  ,
                                                            jstring key) {
    return NULL;
}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    storeRecord
 * Signature: (Ljava/lang/String;[Ljava/lang/Object;)[Ljava/lang/Object;
 */
JNIEXPORT jobjectArray JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_storeRecord(JNIEnv*      env   ,
                                                             jobject      o     ,
                                                             jobject      key   ,
                                                             jobjectArray values) {
    CEPROPVAL  keyValue;
    CEPROPVAL  fields[MAX_FIELDS];
    WORD       nFields;
    jstring    jstr;
    HANDLE     db;
    int        i;

#ifdef DEBUG
    printf("storeRecord(%p, %p, %p, %p)\n", env, o, key, values);
#endif

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return NULL;
    }

    db = getDBHandle(env, o);

    nFields = (WORD)(*env)->GetArrayLength(env, values);

#ifdef DEBUG
    printf("Sort order: %d\n", getKeyField(env, o));
#endif

    //
    // IMPORTANT NOTE: for now we use only string!!!!!!
    // This limitation should change in the future.
    //
    keyValue.propid = MAKELONG(CEVT_LPWSTR, getKeyField(env, o));
    keyValue.wLenData = keyValue.wFlags = 0;
    keyValue.val.lpwstr = (WCHAR*)(*env)->GetStringChars(env, (jstring)key, NULL);

    for (i=0; i<nFields; ++i) {
        jstr = (*env)->GetObjectArrayElement(env, values, i);

        fields[i].propid = MAKELONG(CEVT_LPWSTR, i+1);
		fields[i].wLenData = 0;
		fields[i].wFlags = 0;
		fields[i].val.lpwstr = (jstr != NULL) 
                             ? (WCHAR*)(*env)->GetStringChars(env, jstr, NULL) 
                             : TEXT("");
    }

    if (storeRecord(db, &keyValue, nFields, fields) == 0) {
        (*env)->ThrowNew(env, (*env)->FindClass(env, CLASS_DATA_ACCESS_EXCEPTION), "Error writing the record");
    }

    (*env)->ReleaseStringChars(env, key, keyValue.val.lpwstr);
    for (i=0; i<nFields; ++i) {
        jstr = (*env)->GetObjectArrayElement(env, values, i);

        if (jstr != NULL) {
            (*env)->ReleaseStringChars(env, jstr, fields[i].val.lpwstr);
        }
    }

    return values;
}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    setRecordField
 * Signature: (Ljava/lang/Object;ILjava/lang/Object;)V
 */
JNIEXPORT void JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_setRecordField
(JNIEnv *env, jobject o, jobject key, jint pos, jobject value) {
    CEPROPVAL  keyValue;
    CEPROPVAL  fieldValue;
    HANDLE     db;

#ifdef DEBUG
    printf("setRecordField(%p, %p, %p, %d, %p)\n", env, o, key, pos, value);
#endif

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return;
    }

    db = getDBHandle(env, o);

    //
    // IMPORTANT NOTE: for now we use only string!!!!!!
    // This limitation should change in the future.
    //
    keyValue.propid = MAKELONG(CEVT_LPWSTR, getKeyField(env, o));
    keyValue.wLenData = keyValue.wFlags = 0;
    keyValue.val.lpwstr = (WCHAR*)(*env)->GetStringChars(env, (jstring)key, NULL);

    fieldValue.propid = MAKELONG(CEVT_LPWSTR, pos);
    fieldValue.wLenData = keyValue.wFlags = 0;
    fieldValue.val.lpwstr = (WCHAR*)(*env)->GetStringChars(env, (jstring)value, NULL);

    if (storeRecord(db, &keyValue, 1, &fieldValue) == 0) {
        (*env)->ThrowNew(env, (*env)->FindClass(env, CLASS_DATA_ACCESS_EXCEPTION), "Error writing the record");
    }

    (*env)->ReleaseStringChars(env, key, keyValue.val.lpwstr);
    (*env)->ReleaseStringChars(env, key, fieldValue.val.lpwstr);

}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    deleteRecord
 * Signature: (Ljava/lang/Object;)V
 */
JNIEXPORT void JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_deleteRecord
(JNIEnv *env, jobject o, jobject key){
    HANDLE    db      ;
    CEPROPVAL keyValue;

#ifdef DEBUG
    printf("deleteRecord(%p, %p %p)\n", env, o, key);
#endif

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return;
    }

    db = getDBHandle(env, o);

    keyValue.propid = MAKELONG(CEVT_LPWSTR, getKeyField(env, o));
    keyValue.wLenData = keyValue.wFlags = 0;
    keyValue.val.lpwstr = (WCHAR*)(*env)->GetStringChars(env, (jstring)key, NULL);

#ifdef DEBUG
    wprintf(TEXT("Deleting record %s\n"), keyValue.val.lpwstr);
#endif 
    deleteRecord(db, &keyValue);

    (*env)->ReleaseStringChars(env, key, keyValue.val.lpwstr);
}


/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    getNextRecord
 * Signature: (C)[Ljava/lang/Object;
 */
JNIEXPORT jobjectArray JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_getNextRecord
(JNIEnv *env, jobject o, jchar state) {
    HANDLE    db;
    LPBYTE    fields;
    CEPROPVAL stateValue;
    WCHAR     stateStr[2];
    CEOID     oid;
    WORD      nFields;
    DWORD     len;
    jarray    ret;

#ifdef DEBUG
    printf("getNextRecord(%p, %p, %c)\n", env, o, state);
#endif

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return NULL;
    }

    db = getDBHandle(env, o);

    stateStr[0] = btowc(state);
    stateStr[1] = btowc('\0');

    stateValue.propid = MAKELONG(CEVT_LPWSTR, getStatusField(env, o));
    stateValue.wLenData = stateValue.wFlags = 0;
    stateValue.val.lpwstr = stateStr;

    oid = findRecordByStatus(db, &stateValue);
    if (oid == 0) {
        return NULL;
    }

    fields = NULL;
    len = 0;
    CeReadRecordProps(db, CEDB_ALLOWREALLOC, &nFields, NULL, &fields, &len);

    ret = getRecordValues(env, oid, fields, nFields);

    if (fields != NULL) {
        LocalFree(fields);
    }

    return ret;
}

/*
 * Class:     com_funambol_syncclient_sps_pocketpc_RapiDB
 * Method:    getAllRecords
 * Signature: ()[[Ljava/lang/Object;
 */
JNIEXPORT jobjectArray JNICALL 
Java_com_funambol_syncclient_sps_pocketpc_RapiDB_getAllRecords(JNIEnv* env, 
                                                               jobject o  ) {
    int          i       ;
    int          index   ;
    DWORD        err     ;
    DWORD        len     ;
    int          nRecords;
    CEOID        oid     = 0   ;
    CEGUID       volumeId;
    LPBYTE       buf     = NULL;
    WORD         nFields ;
    HANDLE       db      ;
    jclass       cls     ;
    jobjectArray records = NULL;
    jobjectArray values  = NULL;


#ifdef DEBUG
    printf("getAllRecords()\n");
#endif

    if (!sanityCheck(env, o, CHECK_RAPI|CHECK_DBHANDLE)) {
        //
        // we got an excption
        //
        return NULL;
    }

    db = getDBHandle(env, o);

    //
    // I need to know how many records there are in order to size the array
    //
    getVolOid(env, o, &volumeId);
    nRecords = getRecordCount(&volumeId, getDBOid(env, o));

#ifdef DEBUG
    printf("nRecords: %d\n", nRecords);
#endif

    cls = (*env)->FindClass(env, "java/lang/Object");
    records = (*env)->NewObjectArray(env, nRecords, cls, NULL);

    //
    // Start from the beginning
    //
    CeSeekDatabase(db, CEDB_SEEK_BEGINNING, 0, &index);

    i = 0;
    for (i=0; i<nRecords; ++i) {
        nFields = 0; buf = 0; len = 0;
        oid = CeReadRecordProps(db, CEDB_ALLOWREALLOC, &nFields, NULL, &buf, &len);

        if (oid != 0) {
            values = getRecordValues(env, oid, buf, nFields);
            (*env)->SetObjectArrayElement(env, records, i, values);
        } else {
#ifdef DEBUG
            printf("oid for record number %d is 0!\n", i);
#endif
        }

        // Free allocated memory anyway (even if there was an error)
        if (buf != NULL) {
            LocalFree(buf);
        }

        err = CeGetLastError();
        if ((oid == 0) && (err != 0) && (err != ERROR_NO_MORE_ITEMS)) {
            WCHAR msg[256];
        
            wsprintf(msg, TEXT("Error reading records: %d"), db);
            error(msg);
        }
    }

    return records;
}


/*
 * ----------------------------------------------------------- Private functions
 * 
 */

/*
 * Returns the property dbName
 *
 * @param env the JNI environment
 * @param o the java object
 *
 */
static jstring getDBName(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_DBNAME, FIELD_TYPE_DBNAME);

    if (fid == 0) {
      return NULL;
    }

    return (*env)->GetObjectField(env, o, fid);
}

/*
 * Returns the property volume
 *
 * @param env the JNI environment
 * @param o the java object
 *
 */
static jstring getVolume(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_VOLUME, FIELD_TYPE_VOLUME);

    if (fid == 0) {
      return NULL;
    }

    return (*env)->GetObjectField(env, o, fid);
}

/*
 * Returns the field id of the instance member with the given field name and type or 0 if
 * the specified field does not exist.
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param name the field name
 * @param type the field type (for example Ljava/lang/String;)
 */
static jfieldID getFieldID(JNIEnv *env, jobject o, const char* fieldName, const char* fieldType) {
    jclass    cls = (*env)->GetObjectClass(env, o);
    jfieldID  fid ;
    
    fid = (*env)->GetFieldID(env, cls, fieldName, fieldType);
    if (fid == 0) {
#ifdef DEBUG
        printf("Field %s of type %s not found!\n", fieldName, fieldType);
#endif
    }

    return fid;
}

/*
 * Sets the database property dbHandle.
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param handle the database handle
 */
static void setDBHandle(JNIEnv *env, jobject o, HANDLE handle) {
    jfieldID  fid ;
    
    fid = getFieldID(env, o, FIELD_NAME_DBHANDLE, FIELD_TYPE_DBHANDLE);
    if (fid != 0) {
        (*env)->SetLongField(env, o, fid, (jlong)handle);
    }
}

/*
 * Returns the currently opened database handle
 *
 * @param env JNDI environment
 * @param obj the "this" object
 */
static HANDLE getDBHandle(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_DBHANDLE, FIELD_TYPE_DBHANDLE);

    if (fid == 0) {
      return NULL;
    }

    return (HANDLE)(*env)->GetLongField(env, o, fid);
}


/*
 * Sets the property dbOid.
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param oid the database oid
 */
static void setDBOid(JNIEnv *env, jobject o, CEOID oid) {
    jfieldID  fid ;
    
    fid = getFieldID(env, o, FIELD_NAME_DBOID, FIELD_TYPE_DBOID);
    if (fid != 0) {
        (*env)->SetLongField(env, o, fid, (jlong)oid);
    }
}

/*
 * Returns the currently opened database object id
 *
 * @param env JNDI environment
 * @param obj the "this" object
 */
static CEOID getDBOid(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_DBOID, FIELD_TYPE_DBOID);

    if (fid == 0) {
      return 0;
    }

    return (CEOID)(*env)->GetLongField(env, o, fid);
}

/*
 * Sets the property volOid.
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param guid the volume oid
 */
static void setVolOid(JNIEnv *env, jobject o, PCEGUID guid) {
    jfieldID  fid ;
    jobject   array;
    
    fid = getFieldID(env, o, FIELD_NAME_VOLOID, FIELD_TYPE_VOLOID);
    if (fid != 0) {
        array = newByteArray(env, sizeof(CEGUID), (BYTE*)guid);
        (*env)->SetObjectField(env, o, fid, array);
    }
}

/*
 * Returns the currently mounted volume GUID
 *
 * @param env JNDI environment
 * @param obj the "this" object
 */
static void getVolOid(JNIEnv *env, jobject o, PCEGUID volumeId) {
    jfieldID  fid;
    jobject   array;
    
    fid = getFieldID(env, o, FIELD_NAME_VOLOID, FIELD_TYPE_VOLOID);

    if (fid == 0) {
      return;
    }

    array = (*env)->GetObjectField(env, o, fid);
    (*env)->GetByteArrayRegion(env, array, 0, sizeof(CEGUID), (BYTE*)volumeId);    
}


/*
 * Sets the property keyField
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param keyField the key field position
 */
static void setKeyField(JNIEnv *env, jobject o, int keyField) {
    jfieldID  fid ;
    
    fid = getFieldID(env, o, FIELD_NAME_KEYFIELD, FIELD_TYPE_KEYFIELD);
    if (fid != 0) {
        (*env)->SetIntField(env, o, fid, keyField);
    }
}

/*
 * Returns the value of the property keyField
 *
 * @param env JNDI environment
 * @param obj the "this" object
 */
static int getKeyField(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_KEYFIELD, FIELD_TYPE_KEYFIELD);

    if (fid == 0) {
      return -1;
    }

    return (*env)->GetIntField(env, o, fid);
}

/*
 * Sets the property statusField
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param statusField the key field position
 */
static void setStatusField(JNIEnv *env, jobject o, int statusField) {
    jfieldID  fid ;
    
    fid = getFieldID(env, o, FIELD_NAME_STATUSFIELD, FIELD_TYPE_STATUSFIELD);
    if (fid != 0) {
        (*env)->SetIntField(env, o, fid, statusField);
    }
}

/*
 * Returns the value of the property statusField
 *
 * @param env JNDI environment
 * @param obj the "this" object
 */
static int getStatusField(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_STATUSFIELD, FIELD_TYPE_STATUSFIELD);

    if (fid == 0) {
      return -1;
    }

    return (*env)->GetIntField(env, o, fid);
}

/*
 * Sets the property sortField
 * 
 * @param env JNDI environment
 * @param obj the "this" object
 * @param sortField the key field position
 */
static void setSortField(JNIEnv *env, jobject o, int sortField) {
    jfieldID  fid ;
    
    fid = getFieldID(env, o, FIELD_NAME_SORTFIELD, FIELD_TYPE_SORTFIELD);
    if (fid != 0) {
        (*env)->SetIntField(env, o, fid, sortField);
    }
}

/*
 * Returns the value of the property sortField
 *
 * @param env JNDI environment
 * @param obj the "this" object
 */
static int getSortField(JNIEnv *env, jobject o) {
    jfieldID  fid;
    
    fid = getFieldID(env, o, FIELD_NAME_SORTFIELD, FIELD_TYPE_SORTFIELD);

    if (fid == 0) {
      return -1;
    }

    return (*env)->GetIntField(env, o, fid);
}

/* 
 * Returns a Java Object[] array containing the field values of the record
 * specified by oid.
 *
 * @param env the JNI environment
 * @param oid the record oid
 * @param buf memory buffer where the fileds values are stored
 * @param nFields number of fields
 */
static jobjectArray getRecordValues(JNIEnv *env, CEOID oid, LPBYTE buf, WORD nFields) {
    int          i     ;
    CEPROPVAL*   fields;
    WORD         id    ;
    WORD         type  ;
    jobjectArray ret   ;
    jclass       cls   ;

    fields = (CEPROPVAL *)buf;

#ifdef DEBUG
    printf(">>>>>> Record %d\n", oid);
#endif

    cls = (*env)->FindClass(env, "java/lang/Object");
    ret = (*env)->NewObjectArray(env, nFields, cls, NULL);

    for(i=0; i<nFields; ++i) {
        id   = HIWORD(fields[i].propid);
        type = LOWORD(fields[i].propid);


#ifdef DEBUG
        printField(&fields[i]);
#endif
        (*env)->SetObjectArrayElement(env, ret, i, getFieldValue(env, &fields[i], type));
    }

    return ret;

#ifdef DEBUG
    printf("<<<<<< Record %d\n", id);
#endif
}

/*
 * Returns the appropriate jobject for the given value and type. Supported types are:
 *
 * CE type          Java Type
 * ---------------- -----------------------
 * CEVT_BLOB        byte[]
 * CEVT_BOOL        java.lang.Boolean
 * CEVT_FILETIME    java.util.Date
 * CEVT_I2          java.lang.Integer
 * CEVT_I4          java.lang.Integer
 * CEVT_LPWSTR      java.lang.String
 * CEVT_R8          java.lang.Double
 * CEVT_UI2         java.lang.Long
 * CEVT_UI4         java.lang.Long
 *
 * @param value the value
 * @param type value type
 */
static jobject getFieldValue(JNIEnv *env, CEPROPVAL *field, WORD type) {
    CEVALUNION val = field->val;
    jobject ret = NULL;

#ifdef DEBUG
    printf("type: %d\n", type);
#endif

    switch (type) {
        case CEVT_BLOB:
            ret = newByteArray(env, val.blob.dwCount, val.blob.lpb);
            break;
        case CEVT_BOOL:
            ret = newBoolean(env, (jboolean)val.boolVal);
            break;
        case CEVT_FILETIME:
            ret = newDate(env, &val.filetime);
            break;
        case CEVT_I2:
            ret = newInteger(env, (jint)val.iVal);
            break;
        case CEVT_I4:
            ret = newLong(env, (jlong)val.lVal);
            break;
        case CEVT_LPWSTR:
            ret = newString(env, val.lpwstr);
            break;
        case CEVT_R8:
            ret = newDouble(env, (jdouble)val.dblVal);
            break;
        case CEVT_UI2:
            ret = newLong(env, (jlong)val.uiVal);
            break;
        case CEVT_UI4:
            ret = newLong(env, (jlong)val.lVal);
            break;
        default:
            ret = newString(env, TEXT("Unknown type"));
            break;
    }

    return ret;
}

/*
 * Creates and returns a new byte[] object
 *
 * @param env   JNI environment
 * @param value the initial buffer
 *
 */
static jobject newByteArray(JNIEnv *env, DWORD len, BYTE value[]) {
    jobject ret = NULL;

    ret = (*env)->NewByteArray(env, len);
    (*env)->SetByteArrayRegion(env, ret, 0, len, value);

    return ret;
}

/*
 * Creates and returns a new Boolean
 *
 * @param env   JNI environment
 * @param value the initial boolean value
 *
 */
static jobject newBoolean(JNIEnv *env, jboolean value) {
    jclass    cls = NULL;
    jmethodID mid =    0;

    cls = (*env)->FindClass(env, "java/lang/Boolean");
    mid = (*env)->GetMethodID(env, cls, "<init>", "(Z)V");

    if (mid != 0) {
        return (*env)->NewObject(env, cls, mid, value);
    }
#ifdef DEBUG
    else {
        error(TEXT("constructor <init>(Z)V not found\n"));
    }
#endif

    return NULL;
}

/*
 * Creates and returns a new Integer
 *
 * @param env   JNI environment
 * @param value the initial int value
 *
 */
static jobject newInteger(JNIEnv *env, jint value) {
    jclass    cls = NULL;
    jmethodID mid =    0;

    cls = (*env)->FindClass(env, "java/lang/Integer");
    mid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");

    if (mid != 0) {
        return (*env)->NewObject(env, cls, mid, value);
    }
#ifdef DEBUG
    else {
        error(TEXT("constructor <init>(I)V not found\n"));
    }
#endif
    return NULL;
}

/*
 * Creates and returns a new Long
 *
 * @param env   JNI environment
 * @param value the initial long value
 *
 */
static jobject newLong(JNIEnv *env, jlong value) {
    jclass    cls = NULL;
    jmethodID mid =    0;

    cls = (*env)->FindClass(env, "java/lang/Long");
    mid = (*env)->GetMethodID(env, cls, "<init>", "(J)V");

    if (mid != 0) {
        return (*env)->NewObject(env, cls, mid, value);
    }
#ifdef DEBUG
    else {
        error(TEXT("constructor <init>(J)V not found\n"));
    }
#endif

    return NULL;
}

/*
 * Creates and returns a new Double
 *
 * @param env   JNI environment
 * @param value the initial double value
 *
 */
static jobject newDouble(JNIEnv *env, jdouble value) {
    jclass    cls = NULL;
    jmethodID mid =    0;

    cls = (*env)->FindClass(env, "java/lang/Double");
    mid = (*env)->GetMethodID(env, cls, "<init>", "(D)V");

    if (mid != 0) {
        return (*env)->NewObject(env, cls, mid, value);
    }
#ifdef DEBUG
    else {
        error(TEXT("constructor <init>(Z)V not found\n"));
    }
#endif

    return NULL;
}

/*
 * Creates and returns a new String
 *
 * @param env   JNI environment
 * @param value the initial string value
 *
 */
static jobject newString(JNIEnv *env, WCHAR *value) {
    return (*env)->NewString(env, (jchar*)value, wcslen(value));
}

/*
 * Creates and returns a new java.util.Date object
 *
 * @param env   JNI environment
 * @param value the initial FILETIME value
 *
 */
static jobject newDate(JNIEnv *env, FILETIME* timestamp) {

    jclass    cls = NULL;
    jmethodID mid =    0;

    SYSTEMTIME sysTime;
    struct tm t;
    time_t value;

    //
    // To convert a filetime to a java.util.Date, we first need to 
    // convert it into a system time and then extract the millisenconds
    // in a Java-compatible way. For this, the mktime() C function is
    // used.
    //
    FileTimeToSystemTime(timestamp, &sysTime);

    
    t.tm_year  = sysTime.wYear-1900;
    t.tm_mon   = sysTime.wMonth-1;
    t.tm_mday  = sysTime.wDay;
    t.tm_hour  = sysTime.wHour;
    t.tm_min   = sysTime.wMinute;
    t.tm_sec   = sysTime.wSecond;
    t.tm_wday  = sysTime.wDayOfWeek;
    t.tm_yday  = 0;

#ifdef DEBUG
    {
        char buf[50];

        strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &t);

        sprintf("Creating a date for %s\n", buf);
    }
#endif

    value = mktime(&t);

    cls = (*env)->FindClass(env, "java/util/Date");
    mid = (*env)->GetMethodID(env, cls, "<init>", "(J)V");

    if (mid != 0) {
        return (*env)->NewObject(env, cls, mid, value);
    }
#ifdef DEBUG
    else {
        error(TEXT("constructor <init>(J)V not found\n"));
    }
#endif

    return NULL;
}

/* 
 * Initializes RAPI
 */
static void init() {
    HRESULT res;
    RAPIINIT init;

    init.cbSize = sizeof(RAPIINIT);
	res = CeRapiInitEx(&init);

	if ((res != S_OK) && (init.hrRapiInit != S_OK)) {
#ifdef DEBUG
		error(TEXT("Unable to initialize RAPI"));
#endif
		return;
	}

#ifdef DEBUG
    printf("RAPI initialized\n");
#endif

    initialized = TRUE;
}

/*
 * De-initializes RAPI
 */
static void uninit() {
    CeRapiUninit();

#ifdef DEBUG
    printf("RAPI Finalized\n");
#endif
}

/*
 * Checks that the state of the class is healthy for database access.
 * The following is checked:
 *   - RAPI is initialized
 *   - the property dbHandle is set to a valid handle
 *
 * In case of problems a DataAccessException is thrown.
 *
 * @param env the JNI environment
 * @param o this object
 * @param flags the checks to perform. One of:
 *                  CHECK_RAPI, CHECK_DBHANDLE
 */
static BOOL sanityCheck(JNIEnv *env, jobject o, WORD flags) {
    char msg[256];
    jclass c;

    *msg = 0;
    if ((flags & CHECK_RAPI) && (initialized == FALSE)) {
        strcpy(msg, "RAPI not initialized!");
        goto exc;
    }

    if (flags & CHECK_DBHANDLE) {
        HANDLE db = getDBHandle(env, o);

        if ((db == INVALID_HANDLE_VALUE) || (db == 0)) {
            sprintf(msg, "Invalid database handle: %d", db);
            goto exc;
        }
    }

    return TRUE;

    exc:

        //
        // If msg is not empty, an error occurred
        //
        if (*msg != 0) {
            //
            // It should not happen that CLASS_DATA_ACCESS_EXCEPTION is not
            // found, but sometime it happens. In this case, we do not throw
            // any axception
            //
            c = (*env)->FindClass(env, CLASS_DATA_ACCESS_EXCEPTION);
            if (c != NULL) {
                (*env)->ThrowNew(env, c, msg);
            }
        }

        return FALSE;
}